#! /usr/bin/python

## session.py
## Copyright (C) 2007 Ivan Uemlianin
## Contact: ivan@llaisdy.com
## 
## This file is part of Trefnydd.
##
## Trefnydd is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## Trefnydd is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Trefnydd (i.e., doc/gpl.txt); if not, write to the Free Software
## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
## or see http://www.gnu.org/licenses/gpl.txt


## Parts of this module are based on pyhtk, developed at
## Canolfan Bedwyr and to which the following coyright notice applies:
##
## # Copyright (c) 2005, University of Wales, Bangor 
## # All rights reserved.
## # 
## # Redistribution and use in source and binary forms, with or without
## # modification, are permitted provided that the following conditions
## # are met:
## # 
## #   * Redistributions of source code must retain the above copyright
## # notice, this list of conditions and the following disclaimer.
## #   * Redistributions in binary form must reproduce the above copyright
## # notice, this list of conditions and the following disclaimer in the
## # documentation and/or other materials provided with the distribution.
## #   * Neither the name of the University of Wales, Bangor nor the names
## # of its contributors may be used to endorse or promote products
## # derived from this software without specific prior written permission.
## # 
## # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## # "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## # LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## # A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## # OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## # SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## # LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## # DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## # THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
## For pyhtk see:
## http://www.bangor.ac.uk/~cbs007/pyhtk/README.html
## http://www.bangor.ac.uk/ar/cb/wispr.php
## http://www.e-gymraeg.org/wispr/


"""
========
trefnydd
========

Package:   trefnydd
Module:    lib.htk.session
Copyright: 2007 Ivan A. Uemlianin
Contact:   ivan@llaisdy.com
Licence:   GNU GPL (see doc/gpl.txt)

htk.session
======

"""

import os
from trefnydd.utils import file_IO, location

class HTKMacroSet(dict):
    """** HTK Macros

See htkbook/node104_ct.html Macro Definitions for details.

Name   Description
====   ===========
~c     Shared choleski L' matrix
~d     Shared duration parameters
~h     Physical HMM
~i     Shared inverse full covariance matrix
~j     Linear transform
~l     Logical HMM
~m     Shared Gaussian mixture component
~o     Global option values
~p     Tied mixture
~r     Regression class tree
~s     Shared state distribution
~t     Shared transition matrix
~u     Shared mean vector
~v     Shared diagonal variance vector
~w     Shared stream weight vector
~x     Shared arbitrary transform matrix

    """
    _keys = ['o', 'c', 'd', 'h', 
             'i', 'j', 'l', 'm', 
             'p', 'r', 's', 't',
             'u', 'v', 'w', 'x']
    global_opts_key = _keys[0]

   
    def __init__(self):
        dict.__init__(self)
        for k in self._keys:
            self.setdefault(k, {})

    def __str__(self):
        out = '~%s\n%s\n\n' % (self.global_opts_key,
                               self[self.global_opts_key])
        keys = self._keys
        keys.remove(self.global_opts_key)
        for k in keys:
            if self[k]:
                for var in self[k].keys():
                    out = '%s~%s  %s\n%s\n\n' % (out, k, var, self[k][var])
        return out


class HMM(object):
    def __init__(self):
        pass
    
class HTKHMM(HMM):
    def __init__(self, label=''):
        HMM.__init__(self)
        self.name = label
        self.states = []
        self.transp = None
        # transp: matrix of initial transition probabilities

        # todo: add default states

    def __str__(self):
        out = '~h "%s"\n<BeginHMM>\n<NumStates> %d\n' \
              % (self.name, len(self.states))
        for s in self.states[1:-1]:
            out = '%s%s\n' % (out, s)
        out = '%s<TransP> %d\nTODO' % (out, len(self.states))
        out = '%s\n<EndHMM>\n' % out
        return out

class HMMState(object):
    def __init__(self):
        pass
    
class HTKHMMState(HMMState):
    def __init__(self):
        HMMState.__init__(self)

    def __str__(self):
        out = 'TODO'
        return out
        

prototypeTemplate = """~h "LABEL"
<BeginHMM>
<NumStates> 5
<State> 2
<Mean> 39
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0
<Variance> 39
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0
<State> 3
<Mean> 39
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0
<Variance> 39
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0
<State> 4
<Mean> 39
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0
0.0 0.0 0.0
<Variance> 39
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0
1.0 1.0 1.0
<TransP> 5
0.0 1.0 0.0 0.0 0.0
0.0 0.6 0.4 0.0 0.0
0.0 0.0 0.6 0.4 0.0
0.0 0.0 0.0 0.7 0.3
0.0 0.0 0.0 0.0 0.0
<EndHMM>
"""
    
   

class HTKSession(location.Located):
    config = {'SOURCEKIND': 'WAVEFORM',
              'SOURCEFORMAT': 'WAV',
              'SOURCERATE': 625,
              'TARGETKIND': 'MFCC_0_D_A',
              'TARGETRATE': 100000,
              'SAVECOMPRESSED': 'T',
              'SAVEWITHCRC': 'T',
              'WINDOWSIZE': 250000,
              'USEHAMMING': 'T',
              'PREEMCOEF': 0.97,
              'NUMCHANS': 26,
              'CEPLIFTER': 22,
              'NUMCEPS': 12,
              'ENORMALISE': 'T'}

    def __init__(self, debug=False):
        location.Located.__init__(self, debug)
        self.debug = debug
        self._dirs = {
            'hmm0': 'hmm0',
            'lab': 'lab',
            'log': 'log',
            'mfc': 'mfc',
            'results': 'results',
            'wav': 'wav',
            'gram': 'gram',
            'wdnet': 'wdnet'
            }

        self._fns = {
            'config': 'config',
            'convert_scp': 'convert.scp',
            'hmmlist': 'HmmList',
            'labels_mlf': 'labels.mlf',
            'macros': 'macros',
            'trainlist': 'trainlist',
            'dict': 'dict',
            'hmmdefs': 'hmmdefs'
            }

        self.corpus = None
        self.labelType = None
        self.hmms = set()

    def setup(self, corp=None):
        if corp: self.corpus = corp
        self.make_dirs()
        self.make_mlf()
        self.make_hmmlist()
        self.make_convert_scp()
        self.make_trainlist()
        self.make_config()
        self.make_macros()
        self.make_prototypes()

    def make_dirs(self):
        for dirn in self._dirs:
            self._dirs[dirn] = file_IO.getDir(os.path.join(self.homeDir,
                                                           self._dirs[dirn]))

    def make_mlf(self):
        """Make Master Label File
        """
        out = '#!MLF!#\n'
        for utt in self.corpus:
            out = '%s%s' % (out, utt.write_HTKLab(self.labelType))
        self.export2file(out, 'labels_mlf')
    
    def make_hmmlist(self):
        self.corpus.makeIndex(self.labelType)
        labList = self.corpus.indices[self.labelType].keys()
        labList.sort()
        for lab in labList:
            self.hmms.add(HTKHMM(lab))
        out = '"%s"\n' % '"\n"'.join(labList)
        self.export2file(out, 'hmmlist')

    def make_convert_scp(self):
        """line format: wav/punk1.wav mfc/punk1.mfc"""
        mfcDir = self._dirs['mfc']
        out = ''
        for utt in self.corpus:
            wavfn = utt.audioFn
            mfcfn = os.path.join(mfcDir, ''.join([utt.name, '.mfc']))
            out = '%s%s   %s\n' % (out, wavfn, mfcfn)
        self.export2file(out, 'convert_scp')

    def make_trainlist(self):
        """line format:
        C:\Autosegmentation\HTK\HINIT\TrainingData\mfc\punk1.mfc"""
        mfcDir = self._dirs['mfc']
        out = ''
        for utt in self.corpus:
            mfcfn = os.path.join(mfcDir, ''.join([utt.name, '.mfc']))
            out = '%s%s\n' % (out, mfcfn)
        self.export2file(out, 'trainlist')

    def make_config(self):
        """\
        kHz     SOURCERATE
        =====   =====
        16000   625
        44100   226.757369614512
        """
        # todo/danger: assumes all wavs in corpus have same sample rate
        self.config['SOURCERATE'] = 10000000.0/self.corpus[0].audioSampleRate
        out = '\n'.join(['%s = %s' % (i[0], i[1])
                         for i in self.config.items()])
        out = '# Coding parameters\n%s\n' % out
        self.export2file(out, 'config')
    
    def make_macros(self):
        self.macros = HTKMacroSet()
        self.macros['o'] = """<STREAMINFO> 1 39
<VECSIZE> 39<NULLD><MFCC_0_D_A><DIAGC>"""
        
        self.macros['v']['varFloor1'] = """<Variance> 39
8.713753e-001 6.625333e-001 4.690533e-001 6.571800e-001 4.601170e-001 1.138707e+000 6.161002e-001
5.646990e-001 6.854044e-001 5.973042e-001 4.259520e-001 3.921174e-001 1.099038e+000 3.838177e-002
3.454069e-002 2.658178e-002 3.022736e-002 2.379769e-002 5.005486e-002 3.312984e-002 3.214758e-002
3.607899e-002 3.012902e-002 2.400114e-002 2.118599e-002 3.008951e-002 5.342365e-003 5.642878e-003
4.493490e-003 4.974959e-003 4.013374e-003 7.861852e-003 5.684706e-003 5.699892e-003 6.212835e-003
5.134702e-003 4.360192e-003 3.754829e-003 4.688293e-003
"""
        self.export2file(self.macros.__str__(), 'macros')

    def make_prototypes(self):
        raise StandardError
        for hmm in self.hmms:
            fullFn = os.path.join(self.homeDir,
                                  self._dirs['hmm0'],
                                  hmm.name)
            open(fullFn, 'w').write(hmm.__str__())
    
    def export2file(self, out, fn):
        fullFn = os.path.join(self.homeDir, self._fns[fn])
        open(fullFn, 'w').write(out)
        


def main():
    pass

if __name__ == '__main__':
    main()
 
